# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Base client for authentication and communicating with OSDU."""

import logging
import os

from .base import OsduBaseCredential

logger = logging.getLogger(__name__)


class OsduMsalInteractiveCredential(OsduBaseCredential):
    """Refresh token based client for connecting with OSDU."""

    # __access_token_expire_date = None
    __access_token = None
    # __id_token = None

    @property
    def client_id(self) -> str:
        """Client id used for authorisation

        Returns:
            str: client id
        """
        return self._client_id

    @property
    def authority(self) -> str:
        """Authority url for obtaining token

        Returns:
            str: authority url
        """
        return self._authority

    @property
    def scopes(self) -> str:
        """The current scopes requested

        Returns:
            str: scopes
        """
        return self._scopes

    @property
    def token_cache(self) -> str:
        """Path to persist tokens to

        Returns:
            str: token cache path
        """
        return self._token_cache

    # pylint: disable=too-many-arguments
    def __init__(self, client_id: str, authority: str, scopes: str, token_cache: str = None):
        """Setup the new client

        Args:
            client_id (str): client id for connecting
            authority (str): authority url
            scopes (str): scopes to request
            token_cache (str): path to persist tokens to
        """
        super().__init__()
        self._client_id = client_id
        self._authority = authority
        self._scopes = scopes
        self._token_cache = token_cache

    def get_token(self, **kwargs) -> str:
        """
        Check expiration date and return access_token.
        """
        # if datetime.now().timestamp() > self.__access_token_expire_date:
        self.refresh_access_token()
        return self.__access_token

    def _refresh_access_token(self) -> dict:
        """Refresh token using msal.

        Returns:
            dict: Dictionary representing the returned token
        """
        import msal

        # Create a preferably long-lived app instance which maintains a persistant token cache.
        cache = msal.SerializableTokenCache()
        if os.path.exists(self._token_cache):
            with open(self._token_cache, "r", encoding="utf8") as cachefile:
                cache.deserialize(cachefile.read())

        app = msal.PublicClientApplication(
            self._client_id, authority=self._authority, token_cache=cache
        )

        result = None

        # Firstly, check the cache to see if this end user has signed in before
        # accounts = app.get_accounts(username=config.get("username"))
        accounts = app.get_accounts()
        if accounts:
            logger.debug("Account(s) exists in cache, probably with token too. Let's try.")
            # for a in accounts:
            #     print(a["username"])
            chosen = accounts[
                0
            ]  # Assuming the end user chose this one to proceed - should change if multiple
            # Now let's try to find a token in cache for this account
            result = app.acquire_token_silent([self._scopes], account=chosen)

        if not result:
            logger.debug("No suitable token exists in cache. Let's get a new one from AAD.")
            print("A local browser window will be open for you to sign in. CTRL+C to cancel.")
            result = app.acquire_token_interactive(
                [self._scopes],
                timeout=10,
                # login_hint=config.get("username"),  # Optional.
                # If you know the username ahead of time, this parameter can pre-fill
                # the username (or email address) field of the sign-in page for the user,
                # Often, apps use this parameter during reauthentication,
                # after already extracting the username from an earlier sign-in
                # by using the preferred_username claim from returned id_token_claims.
                # Or simply "select_account" as below - Optional. It forces to show account selector page
                prompt=msal.Prompt.SELECT_ACCOUNT,
            )

            if cache.has_state_changed:
                with open(self.token_cache, "w", encoding="utf8") as cachefile:
                    cachefile.write(cache.serialize())

        return result

    def refresh_access_token(self) -> dict:
        """Refresh from refresh token.

        Returns:
            dict: Dictionary representing the returned token
        """
        # for i in range(self.retries + 1):
        #     # try several times if there any error
        #     try:
        #         result = self._refresh_access_token(
        #             self._token_endpoint, self._refresh_token, self._client_id, self._client_secret)
        #         break
        #     except HTTPError:
        #         if i == self._retries - 1:
        #             # too many errors, raise original exception
        #             raise
        result = self._refresh_access_token()

        if "preferred_username" in result:
            # TO DO: Save username for later login
            pass

        if "access_token" in result:
            # self.__id_token = result["id_token"]
            self.__access_token = result["access_token"]
            # self.__expire_date = datetime.now().timestamp() + result["expires_in"]

            # logger.info("Token is refreshed.")
        else:
            print(result.get("error"))
            print(result.get("error_description"))
            print(result.get("correlation_id"))

        return result  # You may need this when reporting a bug



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""Constants for identity"""


# pylint: disable=too-few-public-methods
class EnvironmentVariables:
    """Environment variable names"""

    CLIENT_ID = "CLIENT_ID"
    CLIENT_SECRET = "CLIENT_SECRET"
    TOKEN_ENDPOINT = "TOKEN_ENDPOINT"
    REFRESH_TOKEN = "REFRESH_TOKEN"
    AUTHORITY = "AUTHORITY"
    SCOPES = "SCOPES"
    TOKEN_CACHE = "TOKEN_CACHE"

    TOKEN_VARS = (CLIENT_ID, CLIENT_SECRET, TOKEN_ENDPOINT, REFRESH_TOKEN)
    MSAL_INTERACTIVE_VARS = (CLIENT_ID, AUTHORITY, SCOPES)



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Base client for working with the OSDU file API."""

from typing import Union

from osdu.client import OsduClient


class ServiceClientBase:
    """Abstract base service client class for connecting with OSDU.
    It is not intended to use this directly, rather one of it's subclasses.
    """

    @property
    def service_name(self) -> str:
        """Get name of the service.
        By default used for constructing the url unless api_url is overwridden.

        Returns:
            str: Name of the service
        """
        return self._service_name

    @property
    def valid_service_versions(self) -> list:
        """List of valid service versions

        Returns:
            list: List of valid service versions
        """
        return self._valid_service_versions

    @property
    def service_version(self) -> int:
        """Version of the api being used

        Returns:
            int: Version of the api being used
        """
        return self._service_version

    def __init__(
        self,
        client: OsduClient,
        service_name: str,
        valid_service_versions: list,
        service_version: Union[int, str] = "latest",
    ):
        """Setup the ServiceClientBase

        Args:
            client (OsduClient): client to use for connection.
            service_name (str): service name.
            valid_service_versions (list): list of valid service versions
            service_version (Union[int, str], optional): service version (3 or 'latest') Defaults to 'latest'.

        Raises:
            ValueError: [description]
        """

        if not client or not isinstance(client, OsduClient):
            raise ValueError("client should be an OsduClient instance")

        if service_version not in valid_service_versions and service_version != "latest":
            raise ValueError(
                f"This package doesn't support API version '{service_version}'.\n"
                + f"Supported versions: {', '.join(str(v) for v in valid_service_versions)} or 'latest'"
            )

        self._client = client
        self._service_name = service_name
        self._valid_service_versions = valid_service_versions
        self._service_version = (
            self.valid_service_versions[-1] if service_version == "latest" else service_version
        )

    def api_url(self, extra_path: str = None):
        """Get a url for the api including any specified extra path

        Args:
            extra_path (str, optional): extra path to add to the base url. Defaults to None.

        Returns:
            str: api url
        """
        url = (
            self._client.server_url.rstrip("/")
            + f"/api/{self.service_name}/v{self.service_version}/"
        )
        if extra_path is not None:
            url = url + extra_path
        return url



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Useful functions."""

import logging
from typing import Union

import requests
from requests.models import HTTPError

from osdu.identity import OsduBaseCredential

logger = logging.getLogger(__name__)


class OsduClient:
    """
    Class for connecting with API's.
    """

    @property
    def server_url(self) -> str:
        """Url of the API server

        Returns:
            str: api server url
        """
        return self._server_url

    @property
    def data_partition(self) -> str:
        """Name of the data partition

        Returns:
            str: data partition name
        """
        return self._data_partition

    @property
    def credentials(self) -> str:
        """Credentials used for connection

        Returns:
            OsduBaseCredential: credentials
        """
        return self._credentials

    @property
    def retries(self) -> int:
        """Number of retries incase of http errors

        Returns:
            int: number of retries incase of http errors
        """
        return self._retries

    def __init__(
        self,
        server_url: str,
        data_partition: str,
        credentials: OsduBaseCredential,
        retries: int = 0,
    ):
        """Setup the new client

        Args:
            server_url (str): url of the server without any path e.g. https://www.test.com
            data_partition (str): data partition name e.g. opendes
            credentials (OsduBaseCredential): credentials used for connection
            retries (int): number of retries incase of http errors (default 0 - no retries)
        """
        self._server_url = server_url
        self._data_partition = data_partition
        self._credentials = credentials
        self._retries = retries

    def get_headers(self):
        """Get needed http headers, including authorization bearer token.

        Raises:
            NotImplementedError: Should be implemented by subclasses.
        """
        return {
            "Content-Type": "application/json",
            "data-partition-id": self.data_partition,
            "Authorization": f"Bearer {self.credentials.get_token()}",
        }

    # region HTTP methods
    def get(self, url: str) -> requests.Response:
        """GET from the specified url

        Args:
            url (str): url to GET from to

        Returns:
            requests.Response: response object
        """
        headers = self.get_headers()
        response = requests.get(url, headers=headers)
        return response

    def get_returning_json(self, url: str, ok_status_codes: list = None) -> dict:
        """Get data from the specified url in json format.

        Args:
            url (str): url to GET from to
            ok_status_codes (list, optional): Status codes for successful call. Defaults to [200].

        Raises:
            HTTPError: Raised if the get returns a status other than those in ok_status_codes

        Returns:
            dict: response json
        """
        if ok_status_codes is None:
            ok_status_codes = [200]
        response = self.get(url)
        if response.status_code not in ok_status_codes:
            raise HTTPError(response=response)
        return response.json()

    def post(self, url: str, data: Union[str, dict]) -> requests.Response:
        """POST data to the specified url

        Args:
            url (str): url to POST to
            data (Union[str, dict]): json data as string or dict to send as the body

        Returns:
            [requests.Response]: response object
        """
        headers = self.get_headers()
        # logger.debug(url)
        # logger.debug(data)

        # determine whether to send to requests as data or json
        _json = None
        if isinstance(data, dict):
            _json = data
            data = None

        response = requests.post(url, data=data, json=_json, headers=headers)
        # logger.debug(response.text)
        return response

    def post_returning_json(
        self, url: str, data: Union[str, dict], ok_status_codes: list = None
    ) -> dict:
        """Post data to the specified url and get the result in json format.

        Args:
            url (str): url to POST to
            data (Union[str, dict]): json data as string or dict to send as the body
            ok_status_codes (list, optional): Status codes indicating successful call. Defaults to [200].

        Raises:
            HTTPError: Raised if the get returns a status other than those in ok_status_codes

        Returns:
            dict: response json
        """
        if ok_status_codes is None:
            ok_status_codes = [200]
        response = self.post(url, data)
        if response.status_code not in ok_status_codes:
            raise HTTPError(response=response)
        return response.json()

    def put(self, url: str, filepath: str) -> requests.Response:
        """PUT from the file at the given path to a url

        Args:
            url (str): url to PUT to
            filepath (str): path to a file to PUT

        Returns:
            requests.Response: response object
        """
        headers = self.get_headers()
        headers.update({"Content-Type": "application/octet-stream", "x-ms-blob-type": "BlockBlob"})
        with open(filepath, "rb") as file_handle:
            response = requests.put(url, data=file_handle, headers=headers)
            return response

    def delete(self, url: str, ok_status_codes: list = None) -> requests.Response:
        """GET to a url

        Args:
            url (str): url to PUT to
            ok_status_codes (list, optional): Status codes indicating successful call. Defaults to [200].

        Returns:
            requests.Response: response object
        """
        if ok_status_codes is None:
            ok_status_codes = [200]

        headers = self.get_headers()
        response = requests.delete(url, headers=headers)

        if response.status_code not in ok_status_codes:
            raise HTTPError(response=response)

        return response

    # endregion HTTP Actions



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Base client for authentication and communicating with OSDU."""

import logging
import os

from osdu.identity.consts import EnvironmentVariables
from osdu.identity.exceptions import CredentialUnavailableError

from .base import OsduBaseCredential
from .msal_interactive import OsduMsalInteractiveCredential
from .token import OsduTokenCredential

_logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class OsduEnvironmentCredential(OsduBaseCredential):
    """A credential configured by environment variables.

    This credential is capable of authenticating using a refresh token, or msal interactive.
    Configuration is attempted in this order, using these environment variables:
    Refresh token:
      - **CLIENT_ID**: client id for connecting
      - **TOKEN_ENDPOINT**: token endpoint for refreshing token
      - **REFRESH_TOKEN**: refresh token
      - **CLIENT_SECRET**: client secret
    Msal interactive:
      - **CLIENT_ID**: client id for connecting
      - **AUTHORITY**: authority url
      - **SCOPES**: scopes to request.
      - **TOKEN_CACHE**: (optional) Path to token cache.

    If a prefix is specified then this is prepended to the above names.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, prefix=None, **kwargs):
        """Setup the new credential based upon environment variables.

        Args:
            prefix (str): optional prefix for standard environment variable names
        """
        self._prefix = prefix
        self._credential = None

        if all(
            os.environ.get(self._expand_environment_name(v)) is not None
            for v in EnvironmentVariables.TOKEN_VARS
        ):
            self._credential = OsduTokenCredential(
                client_id=os.environ[self._expand_environment_name(EnvironmentVariables.CLIENT_ID)],
                token_endpoint=os.environ[
                    self._expand_environment_name(EnvironmentVariables.TOKEN_ENDPOINT)
                ],
                refresh_token=os.environ[
                    self._expand_environment_name(EnvironmentVariables.REFRESH_TOKEN)
                ],
                client_secret=os.environ[
                    self._expand_environment_name(EnvironmentVariables.CLIENT_SECRET)
                ],
                **kwargs,
            )
        elif all(
            os.environ.get(self._expand_environment_name(v)) is not None
            for v in EnvironmentVariables.MSAL_INTERACTIVE_VARS
        ):
            self._credential = OsduMsalInteractiveCredential(
                client_id=os.environ[self._expand_environment_name(EnvironmentVariables.CLIENT_ID)],
                authority=os.environ[self._expand_environment_name(EnvironmentVariables.AUTHORITY)],
                scopes=os.environ[self._expand_environment_name(EnvironmentVariables.SCOPES)],
                token_cache=os.environ.get(
                    self._expand_environment_name(EnvironmentVariables.TOKEN_CACHE)
                ),
                **kwargs,
            )

        if self._credential:
            _logger.info("Environment is configured for %s", self._credential.__class__.__name__)
        else:
            expected_variables = set(
                EnvironmentVariables.TOKEN_VARS + EnvironmentVariables.MSAL_INTERACTIVE_VARS
            )
            set_variables = [v for v in expected_variables if v in os.environ]
            if set_variables:
                _logger.warning(
                    "Incomplete environment configuration. These variables are set: %s",
                    ", ".join(set_variables),
                )
            else:
                _logger.warning("No environment configuration found.")

    def _expand_environment_name(self, name: str):
        return name if self._prefix is None else f"{self._prefix}_{name}"

    def get_token(self, **kwargs) -> str:
        """Get token, deferring to the relevant credential class"""

        if not self._credential:
            message = "OsduEnvironmentCredential unavailable. Environment variables are not fully configured."
            raise CredentialUnavailableError(message=message)
        return self._credential.get_token(**kwargs)



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------

"""OSDU SDK package that can be installed using setuptools"""

# For reasons behind /src/ file structure see https://blog.ionelmc.ro/2014/05/25/python-packaging/#the-structure

import os
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext
from setuptools import setup, find_packages


def read(fname):
    """Local read helper function for long documentation"""
    osdu_path = dirname(os.path.realpath(__file__))
    return open(join(osdu_path, fname)).read()


version_file = read(os.path.join('src', 'osdu', '__init__.py'))
__VERSION__ = re.search(r'^__VERSION__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        version_file, re.MULTILINE).group(1)

setup(
    name='osdu-sdk',
    version=__VERSION__,
    description='OSDU SDK for Python',
    long_description=read('README.rst'),
    url='https://github.com/equinor/osdu-sdk-python',
    author='Equinor ASA',
    author_email='mhew@equinor.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    keywords='osdu sdk python',
    python_requires='>=3.8',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    install_requires=[
        'requests',
        'msal'
    ],
    project_urls={
        'Issue Tracker': 'https://github.com/equinor/osdu-sdk-python/issues',
    }
)



# -----------------------------------------------------------------------------
# Copyright (c) Equinor ASA. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -----------------------------------------------------------------------------
"""Entitlements client for working with the OSDU entitlements API."""

from typing import Union

from osdu.client import OsduClient
from osdu.serviceclientbase import ServiceClientBase

VALID_ENTITLEMENTS_API_VERSIONS = [2]


class EntitlementsClient(ServiceClientBase):
    """A client for working with the OSDU Entitlements API."""

    def __init__(self, client: OsduClient, service_version: Union[int, str] = "latest"):
        """Setup the EntitlementsClient

        Args:
            client (OsduClient): client to use for connection
            service_version (Union[int, str], optional): service version (3 or 'latest') Defaults to 'latest'.

        Raises:
            ValueError: [description]
        """
        super().__init__(client, "entitlements", VALID_ENTITLEMENTS_API_VERSIONS, service_version)

    # def query():
    #     pass

    # def query_by_id():
    #     pass

    def is_healthy(self) -> bool:
        """Returns health status of the API

        Returns:
            bool: health status of the API
        """
        response = self._client.get(self.api_url("health/readiness_check"))
        return response.status_code == 200

    def list_groups(self) -> dict:
        """List groups

        Returns:
            dict: containing the result
        """
        response_json = self._client.get_returning_json(self.api_url("groups"))
        return response_json

    def list_group_members(self, group: str) -> dict:
        """List members in a group

        Args:
            group (str): The email of the group.

        Returns:
            dict: containing the result
        """
        response_json = self._client.get_returning_json(self.api_url(f"groups/{group}/members"))
        return response_json

    def add_group(self, group: str) -> dict:
        """Add a new group

        Args:
            group (str): The email of the group.

        Returns:
            dict: containing the result
        """
        request_data = {"name": group}
        response_json = self._client.post_returning_json(
            self.api_url("groups"), request_data, [200, 201]
        )
        return response_json

    def delete_group(self, group: str):
        """Delete a group

        Args:
            group (str): The email of the group.
        """
        _ = self._client.delete(self.api_url(f"groups/{group}"), [200, 204])

    def add_member_to_group(self, member: str, group: str, role: str) -> dict:
        """Add member to group

        Args:
            member (str): The email of the member to be added.
            group (str): The email of the group.
            role (str): The role in the group.

        Returns:
            dict: containing the result
        """
        request_data = {
            "email": member,
            "role": role,
        }
        response_json = self._client.post_returning_json(
            self.api_url(f"groups/{group}/members"), request_data
        )
        return response_json

    def remove_member_from_group(self, member: str, group: str):
        """Remove member from group

        Args:
            member (str): The email of the member to remove.
            group (str): The email of the group.
        """
        _ = self._client.delete(self.api_url(f"groups/{group}/members/{member}"), [204])
